
import random

USER_AGENTS = [
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.5; en-US; rv:1.9.1b3) Gecko/20090305 Firefox/3.1b3 GTB5",
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.5; ko; rv:1.9.1b2) Gecko/20081201 Firefox/3.1b2",
    "Mozilla/5.0 (X11; U; SunOS sun4u; en-US; rv:1.9b5) Gecko/2008032620 Firefox/3.0b5",
    "Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.8.1.12) Gecko/20080214 Firefox/2.0.0.12",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; cs; rv:1.9.0.8) Gecko/2009032609 Firefox/3.0.8",
    "Mozilla/5.0 (X11; U; OpenBSD i386; en-US; rv:1.8.0.5) Gecko/20060819 Firefox/1.5.0.5",
    "Mozilla/5.0 (Windows; U; Windows NT 5.0; es-ES; rv:1.8.0.3) Gecko/20060426 Firefox/1.5.0.3",
    "Mozilla/5.0 (Windows; U; WinNT4.0; en-US; rv:1.7.9) Gecko/20050711 Firefox/1.0.6",
    "Mozilla/5.0 (Windows; Windows NT 6.1; rv:2.0b2) Gecko/20100720 Firefox/4.0b2",
    "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b4) Gecko/20100818 Firefox/4.0b4",
    "Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.2) Gecko/20100308 Ubuntu/10.04 (lucid) Firefox/3.6 GTB7.1",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:2.0b7) Gecko/20101111 Firefox/4.0b7",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:2.0b8pre) Gecko/20101114 Firefox/4.0b8pre",
    "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b9pre) Gecko/20110111 Firefox/4.0b9pre",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:2.0b9pre) Gecko/20101228 Firefox/4.0b9pre",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:2.2a1pre) Gecko/20110324 Firefox/4.2a1pre",
    "Mozilla/5.0 (X11; U; Linux amd64; rv:5.0) Gecko/20100101 Firefox/5.0 (Debian)",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:6.0a2) Gecko/20110613 Firefox/6.0a2",
    "Mozilla/5.0 (X11; Linux i686 on x86_64; rv:12.0) Gecko/20100101 Firefox/12.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:15.0) Gecko/20120716 Firefox/15.0a2",
    "Mozilla/5.0 (X11; Ubuntu; Linux armv7l; rv:17.0) Gecko/20100101 Firefox/17.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:21.0) Gecko/20130328 Firefox/21.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:22.0) Gecko/20130328 Firefox/22.0",
    "Mozilla/5.0 (Windows NT 5.1; rv:25.0) Gecko/20100101 Firefox/25.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:25.0) Gecko/20100101 Firefox/25.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:28.0) Gecko/20100101 Firefox/28.0",
    "Mozilla/5.0 (X11; Linux i686; rv:30.0) Gecko/20100101 Firefox/30.0",
    "Mozilla/5.0 (Windows NT 5.1; rv:31.0) Gecko/20100101 Firefox/31.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:33.0) Gecko/20100101 Firefox/33.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:40.0) Gecko/20100101 Firefox/40.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.12; rv:58.0) Gecko/20100101 Firefox/58.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/525.19 (KHTML, like Gecko) Chrome/1.0.154.53 Safari/525.19",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/525.19 (KHTML, like Gecko) Chrome/1.0.154.36 Safari/525.19",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.10 (KHTML, like Gecko) Chrome/7.0.540.0 Safari/534.10",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/534.4 (KHTML, like Gecko) Chrome/6.0.481.0 Safari/534.4",
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X; en-US) AppleWebKit/533.4 (KHTML, like Gecko) Chrome/5.0.375.86 Safari/533.4",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/532.2 (KHTML, like Gecko) Chrome/4.0.223.3 Safari/532.2",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/532.0 (KHTML, like Gecko) Chrome/4.0.201.1 Safari/532.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/532.0 (KHTML, like Gecko) Chrome/3.0.195.27 Safari/532.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/530.5 (KHTML, like Gecko) Chrome/2.0.173.1 Safari/530.5",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/534.10 (KHTML, like Gecko) Chrome/8.0.558.0 Safari/534.10",
    "Mozilla/5.0 (X11; U; Linux x86_64; en-US) AppleWebKit/540.0 (KHTML,like Gecko) Chrome/9.1.0.0 Safari/540.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/534.14 (KHTML, like Gecko) Chrome/9.0.600.0 Safari/534.14",
    "Mozilla/5.0 (X11; U; Windows NT 6; en-US) AppleWebKit/534.12 (KHTML, like Gecko) Chrome/9.0.587.0 Safari/534.12",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.13 (KHTML, like Gecko) Chrome/9.0.597.0 Safari/534.13",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.16 (KHTML, like Gecko) Chrome/10.0.648.11 Safari/534.16",
    "Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US) AppleWebKit/534.20 (KHTML, like Gecko) Chrome/11.0.672.2 Safari/534.20",
    "Mozilla/5.0 (Windows NT 6.0) AppleWebKit/535.1 (KHTML, like Gecko) Chrome/14.0.792.0 Safari/535.1",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/535.2 (KHTML, like Gecko) Chrome/15.0.872.0 Safari/535.2",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/535.7 (KHTML, like Gecko) Chrome/16.0.912.36 Safari/535.7",
    "Mozilla/5.0 (Windows NT 6.0; WOW64) AppleWebKit/535.11 (KHTML, like Gecko) Chrome/17.0.963.66 Safari/535.11",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_8) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.45 Safari/535.19",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/535.24 (KHTML, like Gecko) Chrome/19.0.1055.1 Safari/535.24",
    "Mozilla/5.0 (Windows NT 6.2) AppleWebKit/536.6 (KHTML, like Gecko) Chrome/20.0.1090.0 Safari/536.6",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.1 (KHTML, like Gecko) Chrome/22.0.1207.1 Safari/537.1",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.15 (KHTML, like Gecko) Chrome/24.0.1295.0 Safari/537.15",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.93 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/28.0.1467.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/30.0.1599.101 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1623.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1847.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/37.0.2062.103 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.38 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2490.71 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.62 Safari/537.36"
]


SITE = {
        
        "be" : {"site": "https://www.morningstar.be/be/", "iso3" : 'BEL'},
        "br" : {"site": "https://www.morningstarbr.com/br/", "iso3" : 'BRA'},
        "ca": { "site":"https://www.morningstar.ca/ca/", "iso3" : 'CAN'},
        "ch" : { "site":"https://www.morningstar.ch/ch/",  "iso3" : 'CHE'},
        "cl" : {"site": "https://www.morningstar.cl/cl/", "iso3" : 'CHL'},
        "de" : {"site":"https://www.morningstar.de/de/", "iso3" : 'DEU'},
        "dk" : {"site": "https://www.morningstar.dk/dk/", "iso3" : 'DNK'},
        "es" : {"site": "https://www.morningstar.es/es/", "iso3" : 'ESP'},
        "fi" : {"site": "https://www.morningstar.fi/fi/", "iso3" : 'FIN'},
        "fr" : {"site": "https://www.morningstar.fr/fr/", "iso3" : 'FRA'},
        "hk" : {"site": "https://www.morningstar.hk/hk/", "iso3" : 'HKG'},
        "ie" : {"site": "https://www.morningstarfunds.ie/ie/", "iso3" : 'IRL'},
        "it" : {"site": "https://www.morningstar.it/it/", "iso3" : 'ITA'},
        "mx" : {"site": "https://www.morningstar.com.mx/mx/", "iso3" : 'MEX'},
        "nl" : {"site": "https://www.morningstar.nl/nl/", "iso3" : 'NLD'},
        "no" : {"site": "https://www.morningstar.no/no/", "iso3" : 'NOR'},
        "pt" : {"site": "https://www.morningstar.pt/pt/", "iso3" : 'PRT'},
        "se" : {"site": "https://www.morningstar.se/se/", "iso3" : 'SWE'},
        "sg" : {"site": "https://sg.morningstar.com/sg/", "iso3" : 'SGP'}, 
        "uk" : {"site": "https://www.morningstar.co.uk/uk/", "iso3" : "GBR"},
 }




def random_user_agent():
    """
    This function selects a random User-Agent from the User-Agent list, . User-Agents are used in
    order to avoid the limitations of the requests to morningstar.com. The User-Agent is
    specified on the headers of the requests and is different for every request.

   

    Returns:
        :obj:`str` - user_agent:
            The returned :obj:`str` is the name of a random User-Agent, which will be passed on the 
            headers of a request so to avoid restrictions due to the use of multiple requests from the 
            same User-Agent.
    
    """

    return random.choice(USER_AGENTS)



FIELDS = [
    'administratorCompanyId',
    'AlphaM36',
    'AnalystRatingScale',
    'AverageCreditQualityCode',
    'AverageMarketCapital',
    'BetaM36',
    'BondStyleBox',
    'brandingCompanyId',
    'categoryId',
    'CategoryName',
    'ClosePrice',
    'distribution',
    'EffectiveDuration',
    'equityStyle',
    'EquityStyleBox',
    'ExpertiseAdvanced',
    'ExpertiseBasic',
    'ExpertiseInformed',
    'FeeLevel',
    'fundSize',
    'fundStyle',
    'FundTNAV',
    'GBRReturnD1',
    'GBRReturnM0',
    'GBRReturnM1',
    'GBRReturnM12',
    'GBRReturnM120',
    'GBRReturnM3',
    'GBRReturnM36',
    'GBRReturnM6',
    'GBRReturnM60',
    'GBRReturnW1',
    'geoRegion',
    'globalAssetClassId',
    'globalCategoryId',
    'iMASectorId',
    'InitialPurchase',
    'instrumentName',
    'investment',
    'investmentExpertise',
    'investmentExpertise',
    'investmentObjective',
    'investorType',
    'InvestorTypeEligibleCounterparty',
    'InvestorTypeProfessional',
    'InvestorTypeRetail',
    'largestSector',
    'LegalName',
    'managementStyle',
    'ManagerTenure',
    'MaxDeferredLoad',
    'MaxFrontEndLoad',
    'MaximumExitCostAcquired',
    'MorningstarRiskM255',
    'Name',
    'ongoingCharge',
    'OngoingCostActual',
    'PerformanceFeeActual',
    'PriceCurrency',
    'QuantitativeRating',
    'R2M36',
    'ReturnD1',
    'ReturnM0',
    'ReturnM1',
    'ReturnM12',
    'ReturnM120',
    'ReturnM3',
    'ReturnM36',
    'ReturnM6',
    'ReturnM60',
    'ReturnProfileGrowth',
    'ReturnProfileHedging',
    'ReturnProfileIncome',
    'ReturnProfileOther',
    'ReturnProfilePreservation',
    'ReturnW1',
    'riskSrri',
    'SecId',
    'shareClassType',
    'SharpeM36',
    'StandardDeviationM36',
    'starRating',
    'StarRatingM255',
    'SustainabilityRank',
    'sustainabilityRating',
    'TenforeId',
    'totalReturn',
    'totalReturnTimeFrame',
    'TrackRecordExtension',
    'TransactionFeeActual',
    'umbrellaCompanyId',
    'Yield_M12',
    'yieldPercent',
]




import requests
import json
from bs4 import BeautifulSoup

from .utils import random_user_agent
from .utils import SITE




def get_bearer_token():
        """
        This function will scrape the Bearer Token needed to access MS API

        Returns:
        str bearer token
        """
        url = 'https://www.morningstar.co.uk/Common/funds/snapshot/PortfolioSAL.aspx'
        
        headers = {'user-agent' : random_user_agent()}


        response = requests.get(url, headers=headers)
        soup = BeautifulSoup(response.text, 'html.parser')
        script = soup.find('script', {'type':'text/javascript'})
        bearerToken = str(script).split('tokenMaaS:')[-1].split('}')[0].replace('"','').strip()
        return bearerToken

def search_funds(term, field, country ='uk', pageSize=10, currency ='EUR'):
    """
    This function will use the screener of morningstar.co.uk to find funds which include the term.

    Args:
      term (str): text to find a funds can be a the name, part of a name or the isin of the funds
      field (str) : field to find
      iso (str) : text for code ISO 3166-1 alpha-2 of country
      pageSize (int): number of funds to return

    Returns:
      list of dict with SecId, TenforId and LegalName
        [{'SecId': 'F00000270E', 'TenforeId': '52.8.FR0010342600', 'LegalName': '21 Gestion Active'}, {'SecId': 'F000013BGI', 'TenforeId': '52.8.MT7000022612', 'LegalName': '24 Capital Management SICAV plc - 24 Global Currency Fund Share Class A USD Accumulation'}, {'SecId': 'F00000PZHI', 'TenforeId': '52.8.FR0011443225', 'LegalName': '29 Haussmann Actions Europe C'}, {'SecId': 'F0GBR06QS1', 'TenforeId': '52.8.FR0007057427', 'LegalName': '29 Haussmann Actions Europe D'}, {'SecId': 'F0000101BL', 'TenforeId': '52.8.FR0013266590', 'LegalName': '29 Haussmann Actions Europe 
        I'}, {'SecId': 'F00000JW7U', 'TenforeId': '52.8.LU0532306957', 'LegalName': '3F Generation Acc'}, {'SecId': 'F00000UDVR', 'TenforeId': '52.8.FR0011911189', 'LegalName': 'AAM Family Values E1'}, {'SecId': 'F00000UDVS', 'TenforeId': '52.8.FR0011911197', 'LegalName': 'AAM Family Values 
        I'}, {'SecId': 'F0GBR04RG5', 'TenforeId': '52.8.FR0007022025', 'LegalName': 'AAZ Capitalisation'}, {'SecId': 'F000000ITD', 'TenforeId': '52.8.FR0010361600', 'LegalName': 'AAZ Prestige Or'}]
      
      str if the screener find no funds which match the term
        '0 funds found whith the term rzt'
        

    Example:
      >>> search_funds("Myria",country="fr", pageSize=25)
      >>> search_funds("FR0011399914", country="fr", pageSize=25)
      
    
    """

    if not isinstance(field, (str, list)):
      raise TypeError('field parameter should be a string or a list')

    if not isinstance(country, str):
      raise TypeError('country parameter should be a string')

    if not country.lower() in SITE.keys():
      raise ValueError(f'country parameter can only take one of the values: {", ".join(SITE.keys())}')

    if not isinstance(pageSize, int):
      raise TypeError('pageSize parameter should be an integer')


    if isinstance(field, list):
        securityDataPoints = '|'.join(field)
    else:
        securityDataPoints = field

    iso = SITE[country.lower()]["iso3"]


    #url
    url = "https://tools.morningstar.co.uk/api/rest.svc/klr5zyak8x/security/screener"
  
    params = {
    'page' : 1,
    'pageSize' : pageSize,
    'sortOrder' : 'LegalName asc',
    'outputType' : 'json',
    'version' : 1,
    'universeIds' : f'FO{iso}$$ALL',
    'currencyId': currency,
    'securityDataPoints' : securityDataPoints,
    'term' : term,
    }

    #headers
    headers = {
                'user-agent': random_user_agent(),
                }   

    response = requests.get(url,params=params, headers=headers)
    if response.status_code == 200:
            
      result =json.loads(response.content.decode())

      if result['rows']:
        return result['rows']
      else:
        print('0 funds found whith the term %s' % (term))
        return {}

    else:
      raise ConnectionError(f"Response with code {response.status_code}")
    
        




import requests
import json
import re
from bs4 import BeautifulSoup
import pandas as pd

from .utils import random_user_agent
from .utils import SITE, FIELDS
from .search import search_funds, get_bearer_token



class MS:

    def __init__(self, term = None, country: str = "uk", pageSize : int =1, itemRange: int = 0):

        """
        Main class to access MS Data

        Args:
        term (str): text to find a funds can be a the name, part of a name or the isin of the funds
        country (str) : text for code ISO 3166-1 alpha-2 of country
        pageSize (int): number of funds to return
        itemRange (int) : index of funds to return (must be inferior to PageSize)

        Examples:
            >>> MS('0P0000712R', "ca", 9, 0)
            >>> MS('bond', "uk", 25, 2)

        """
        
        
        if not isinstance(country, str):
            raise TypeError('country parameter should be a string')

        if not country.lower() in SITE.keys():
            raise ValueError(f'country parameter can only take one of the values: {", ".join(SITE.keys())}')

        if not isinstance(pageSize, int):
            raise TypeError('pageSize parameter should be an integer')

        if not isinstance(itemRange, int):
            raise TypeError('itemRange parameter should be an integer')

        if pageSize <= itemRange :
            raise ValueError('itemRange parameter should be strictly inferior to pageSize parameter')


        self.site = SITE[country.lower()]["site"]
        
        self.iso3 = SITE[country.lower()]["iso3"]

        self.country = country
        
        code_list = search_funds(term,['SecId','TenforeId','LegalName'], country, pageSize)

        
        if code_list:
            if itemRange < len(code_list):
                self.code = code_list[itemRange]["SecId"]
                self.name = code_list[itemRange]["LegalName"]
                if "TenforeId" in code_list[itemRange]:
                    self.isin = code_list[itemRange]["TenforeId"][-12:]
                else:
                    self.isin = None
            else:
                raise ValueError(f'Found only {len(code_list)} funds found with the term {term}. The paramater itemRange must maximum equal to {len(code_list)-1}')
        else:
            raise ValueError(f'0 funds found with the term {term}')
        


    
    def carbonMetrics(self):
        """
        Returns:
        dict with Carbon Metrics
        """
        

        
        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/esg/carbonMetrics/%s/data""" % (self.code)

        #params
        params = {
                "languageId": "en-GB",
                "locale": "en-GB",
                "clientId": "MDC_intl",
                "benchmarkId": "mstarorcat",
                "component": "sal-components-mip-carbon-metrics",
                "version": "3.60.0"
        }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'en-EN,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
            
            result =json.loads(response.content.decode()) 
        else:
            print("carbonMetrics",response)
            result = {}
        return result

    def esgData(self):
        """        
        Returns:
        dict with ESG Data
        
        """

        
        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/esg/v1/%s/data""" % (self.code)

        #params
        params = {
                "languageId": "en-GB",
                "locale": "en-GB",
                "clientId": "MDC_intl",
                "benchmarkId": "mstarorcat",
                "component": "sal-components-sustainability",
                "version": "3.60.0"
        }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
            
            result =json.loads(response.content.decode()) 
        else:
            print("esgData",response)
            result = {}
        return result

    def creditQuality(self):
        """        
        Returns:
        dict with credit quality data
        
        """


        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/portfolio/creditQuality/%s/data""" % (self.code)

        #params
        params = {
                "languageId": "en-GB",
                "locale": "en-GB",
                "clientId": "MDC_intl",
                "benchmarkId": "mstarorcat",
                "component": "sal-components-mip-credit-quality",
                "version": "3.60.0"
        }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
            
            result =json.loads(response.content.decode()) 
        else:
            print("creditQuality", response)
            result = {}
        return result

    def fixedIncomeStyle(self):
        """        
        Returns:
        dict with credit fixed income data
        
        """

        
        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/process/fixedIncomeStyle/%s/data""" % (self.code)

        #params
        params = {
                "languageId": "en-GB",
                "locale": "en-GB",
                "clientId": "MDC_intl",
                "benchmarkId": "mstarorcat",
                "component": "sal-components-mip-fixed-income-style",
                "version": "3.60.0"
        }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
            
            result =json.loads(response.content.decode()) 
        else:
            print("fixedIncomeStyle", response)
            result = {}
        return result

    def sector(self):
        """        
        Returns:
        dict with sector data
        
        """

        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/portfolio/v2/sector/%s/data""" % (self.code)

        #params
        params = {
                "languageId": "en-GB",
                "locale": "en-GB",
                "clientId": "MDC_intl",
                "benchmarkId": "mstarorcat",
                "component": "sal-components-mip-sector-exposure",
                "version": "3.60.0"
        }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
            
            result =json.loads(response.content.decode()) 
        else:
            print("sector",response)
            result = {}
        return result

    def financialMetrics(self):
        """        
        Returns:
        dict with financial ratio
        
        """

        
        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/process/financialMetrics/%s/data""" % (self.code)

        #params
        params = {
                "languageId": "en-GB",
                "locale": "en-GB",
                "clientId": "MDC_intl",
                "benchmarkId": "mstarorcat",
                "component": "sal-components-mip-financial-metrics",
                "version": "3.60.0"
        }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
            
            result =json.loads(response.content.decode()) 
        else:
            print("financialMetrics",response)
            result = {}
        return result

    def marketCapitalization(self):
        """        
        Returns:
        dict with market capitalization
        
        """

        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/process/marketCap/%s/data""" % (self.code)

        #params
        params = {
                "languageId": "en-GB",
                "locale": "en-GB",
                "clientId": "MDC_intl",
                "benchmarkId": "mstarorcat",
                "component": "sal-components-mip-market-cap",
                "version": "3.60.0"
        }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
        
            result =json.loads(response.content.decode()) 
        else:
            print("marketCapitalization", response)
            result = {}
        return result

    def stockStyle(self):
        """        
        Returns:
        dict with stocks data
        
        """
        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/process/stockStyle/v2/%s/data""" % (self.code)

        #params
        params = {
                "languageId": "en-GB",
                "locale": "en-GB",
                "clientId": "MDC_intl",
                "benchmarkId": "mstarorcat",
                "component": "sal-components-mip-measures",
                "version": "3.60.0"
        }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
        
            result =json.loads(response.content.decode()) 
        else:
            print("stockStyle", response)
            result = {}
        return result

    def factorProfile(self):
        """        
        Returns:
        dict factor profile
        
        """
        
        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/factorProfile/%s/data""" % (self.code)

        #params
        params = {
                "languageId": "en-GB",
                "locale": "en-GB",
                "clientId": "MDC_intl",
                "benchmarkId": "mstarorcat",
                "component": "sal-components-mip-factor-profile",
                "version": "3.60.0"
        }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
        
            result = json.loads(response.content.decode()) 
        else:
            print("factorProfile", response)
            result = {}
        return result

    def ownershipZone(self):
        """        
        Returns:
        dict ownership zone
        
        """
        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/process/ownershipZone/%s/data""" % (self.code)

        #params
        params = {
                "languageId": "en-GB",
                "locale": "en-GB",
                "clientId": "MDC_intl",
                "benchmarkId": "mstarorcat",
                "component": "sal-components-mip-stock-style",
                "version": "3.60.0"
        }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
        
            result =json.loads(response.content.decode())
        else:
            print("get_ownershipZone", response)
            result = {}
        return result

    def assetAllocation(self):
        """        
        Returns:
        dict asset allocation
        
        """

        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/process/asset/v2/%s/data""" % (self.code)

        #params
        params = {
                "languageId": "en-GB",
                "locale": "en-GB",
                "clientId": "MDC_intl",
                "benchmarkId": "mstarorcat",
                "component": "sal-components-mip-asset-allocation",
                "version": "3.60.0"
        }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
            result = json.loads(response.content.decode())
        else:
            print("assetAllocation", response)
            result = {}
        return result



    def holdings(self, holdingType: str = 'all'):
        """        
        Returns:
        pandas DataFrame with funds holdings
        
        """
        holdingType_to_holdingPage = {"all" : "all", "bond" : "boldHoldingPage","equity" : "equityHoldingPage", "other": "otherHoldingPage"}
        if holdingType not in holdingType_to_holdingPage:
            raise ValueError(f'parameter holdingType must take one of the following value : {", ".join(holdingType_to_holdingPage.keys())}  ')
    
        if holdingType == "all":
            return pd.DataFrame(self.position()["equityHoldingPage"]["holdingList"] + 
            self.position()["boldHoldingPage"]["holdingList"] + self.position()["otherHoldingPage"]["holdingList"] )
        else:
           return pd.DataFrame(self.position()[holdingType_to_holdingPage[holdingType]]["holdingList"])



    
    def position(self):
        """        
        Returns:
        dict with positions
        
        """

        
        #url of API
        url = """https://www.us-api.morningstar.com/sal/sal-service/fund/portfolio/holding/v2/%s/data""" % (self.code)

        #params
        params = {
            'premiumNum': '1000',
            'freeNum': '1000',
            'languageId': 'en-GB',
            'locale': 'en-GB',
            'clientId': 'MDC_intl',
            'benchmarkId': 'category',
            'component': 'sal-components-mip-holdings',
            'version': '3.40.1'
            }

        #headers
        headers = {
                    'accept': '*/*',
                    'accept-encoding': 'gzip, deflate, br',
                    'accept-language': 'fr-FR,fr;q=0.9,en-US;q=0.8,en;q=0.7',
                    'authorization': 'Bearer %s' % (get_bearer_token()),
                    'credentials': 'omit',
                    'origin': 'https://www.morningstar.co.uk',
                    'referer': 'https://www.morningstar.co.uk/',
                    'sec-ch-ua': '"Chromium";v="88", "Google Chrome";v="88", ";Not A Brand";v="99"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-fetch-dest': 'empty',
                    'sec-fetch-mode': 'cors',
                    'sec-fetch-site': 'cross-site',
                    'user-agent': random_user_agent(),
                    'X-API-RequestId': '6c20ob56-8055-aa34-9632-5f5a0dbbvboi',
                    'X-SAL-ContentType': 'e7FDDltrTy+tA2HnLovvGL0LFMwT+KkEptGju5wXVTU='
                    }    
        response = requests.get(url,params=params, headers=headers)
        if response.status_code == 200:
            result =json.loads(response.content.decode())
            
        else:
            print("position", response)
            result = {}
        return result


    def objectiveInvestment(self):
        """        
        Returns:
        str objective investment
        
        """
        
        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #Page 1 - overview
        #url page overview
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}"
        #get HTML page overview
        response = requests.get(url, headers=headers)
        #if page not found
        
        if response.status_code != 200:
            raise ConnectionError(f"error {response.status_code}")
        else:
            #html page as soup
            soup = BeautifulSoup(response.text, 'html.parser')
            #investment objective funds
            return soup.find(id='overviewObjectiveDiv').find('td', {"class": "value text"}).text

    
    def referenceIndex(self, index):
        """
        Args:
            index (str) : possible values are benchmark, category     
        Returns:
            str category or benchmark

        Examples:
            >>> MS.referenceIndex("category")
            >>> MS.referenceIndex("benchmark")
        
        """
        
        index_row = {'benchmark' : 0,'category' : 1}
        if index not in index_row:
            raise ValueError(f"index parameter must take one of the following value : { ', '.join(index_row.keys())}")

                #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #Page 1 - overview
        #url page overview
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}"
        #get HTML page overview
        response = requests.get(url, headers=headers)
        #if page not found
        
        if response.status_code != 200:
            raise ConnectionError(f"error {response.status_code}")
        else:
            #html page as soup
            soup = BeautifulSoup(response.text, 'html.parser')
            benchmark_soup = soup.find(id='overviewBenchmarkDiv2Cols').find_all('td', {"class": "value text"})
            return benchmark_soup[index_row[index]].text


    def benchmark(self):
        """
        Returns:
            str benchmark
        """
        return self.referenceIndex("benchmark")


    def category(self):
        """
        Returns:
            str benchmark
        """
        return self.referenceIndex("category")

    def fundsAnnualPerformance(self):
        """
        Returns:
            dict funds annual performance
        """
        return self.AnnualPerformance('funds')


    def categoryAnnualPerformance(self):
        """
        Returns:
            dict category annual performance
        """
        return self.AnnualPerformance('category')


    def indexAnnualPerformance(self):
        """
        Returns:
            dict index annual performance
        """
        return self.AnnualPerformance('index')

    def fundsAnnualRank(self):
        """
        Returns:
            dict funds annual rank
        """
        return self.AnnualPerformance('rank')

    def AnnualPerformance(self, cat):
        """
        Args:
            cat (str) : possible values are category, funds, index, rank    
        Returns:
            dict annual performance or rank

        Examples:
            >>> MS.AnnualPerformance("category")
            >>> MS.AnnualPerformance("funds")
            >>> MS.AnnualPerformance("index")
            >>> MS.AnnualPerformance("rank")
        
        """

        cat_row = {'funds' : 0,'category' : 1, 'index' : 2, 'rank' : 3}
        if cat not in cat_row:
            raise ValueError(f"cat parameter must take one of the following value : { ', '.join(cat_row.keys())}")

        result = {}
        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #page 1 - performance
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=1"
        
        response = requests.get(url, headers=headers)
        soup = BeautifulSoup(response.text, 'html.parser')
        #label are dates
        regex = re.compile('.*heading number')
        label_list = soup.find(id='returnsCalenderYearDiv').find_all('td', {"class": regex})
        #funds performance, category performance, index performance, rank in category
        regex = re.compile('.*value number')
        #values
        value_list = soup.find(id='returnsCalenderYearDiv').find_all('td', {"class": regex})

        regex = re.compile('-|\/')
        #first col is nothing
        for i in range(1, len(label_list)):
            label = label_list[i].text
            #if today

            if regex.search(label):
                label = 'current'
            #add category to label
            if label:
                
                result[f'{cat}_annual_performance_{label}'] = value_list[i+(cat_row[cat])*(len(label_list)-1)-1].text

        return result

    def fundsCumulativePerformance(self):
        """
        Returns:
            dict funds cumulative performance
        """
        return self.CumulativePerformance('funds')

    def categoryCumulativePerformance(self):
        """
        Returns:
            dict category cumulative performance
        """
        
        return self.CumulativePerformance('category')

    def indexCumulativePerformance(self):
        """
        Returns:
            dict index cumulative performance
        """
        
        return self.CumulativePerformance('index')

    def CumulativePerformance(self, cat):
        """
        Args:
            cat (str) : possible values are category, funds, index   
        Returns:
            dict cumulative performance

        Examples:
            >>> MS.CumulativePerformance("category")
            >>> MS.CumulativePerformance("funds")
            >>> MS.CumulativePerformance("index")
        
        """

        cat_row = {'funds' : 2,'category' : 3, 'index' : 4}
        if cat not in cat_row:
            raise ValueError(f"cat parameter must take one of the following value : { ', '.join(cat_row.keys())}")
        result = {}

        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #page 1 - performance
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=1"
        
        response = requests.get(url, headers=headers)
        soup = BeautifulSoup(response.text, 'html.parser')
        cumulative_performance_date = soup.find(id='returnsTrailingDiv').find('td', {"class": "titleBarNote"}).text
        result['cumulative_performance_date'] = cumulative_performance_date 
        #days
        regex = re.compile('.*label')
        label_list = soup.find(id='returnsTrailingDiv').find_all('td', {"class": regex})

        #cumulative performance cat
        regex = re.compile(f'.*col{str(cat_row[cat])} value number')
        value_list = soup.find(id='returnsTrailingDiv').find_all('td', {"class": regex})
        #loop on label
        for i in range(0, len(label_list)):
            #label
            label = label_list[i].text
            #perf funds
            result[f'{cat}_cumulative_performance_{label}'] = re.sub('[^0-9,-\.]','',value_list[i].text)
        return result


    def fundsQuarterlyPerformance(self):
        """
        Returns:
            dict funds quarterly performance
        """
        result = {}
        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #page 1 - performance
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=1"
        
        response = requests.get(url, headers=headers)
        soup = BeautifulSoup(response.text, 'html.parser')
        quarterly_performance_date = soup.find(id='returnsTrailingDiv').find('td', {"class": "titleBarNote"}).text
        result['quarterly_performance_date'] = quarterly_performance_date 

        #quarter label
        regex = re.compile('.*heading number')
        quarter_list = soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #year label
        regex = re.compile('.*label')
        year_list =soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #1st Quarter 
        regex = re.compile('.*col2 value number')
        quarter_1_list = soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #2nd Quarter
        regex = re.compile('.*col3 value number')
        quarter_2_list = soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #3rd Quarter
        regex = re.compile('.*col4 value number')
        quarter_3_list = soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #4th Quarter
        regex = re.compile('.*col5 value number')
        quarter_4_list = soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #loop on year
        for i in range(0, len(year_list)):
            label = 'performance_%s_' % (year_list[i].text)
            result[label + 'quarter_1'] = quarter_1_list[i].text
            result[label + 'quarter_2'] = quarter_2_list[i].text
            result[label + 'quarter_3'] = quarter_3_list[i].text
            result[label + 'quarter_4'] = quarter_4_list[i].text
        return result

    def contact(self):
        """
        Returns:
            dict funds contact like phone number, asset manager
        """
        result = {}
        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #page 1 - performance
        #page 4 - info about found
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=4"
        response = requests.get(url, headers=headers)
        soup = BeautifulSoup(response.text, 'html.parser')
        #label
        label_list = soup.find(id='managementManagementDiv').find_all('td', {"class": "col1 label"})
        #value
        value_list = soup.find(id='managementManagementDiv').find_all('td', {"class": "col2 value number"})
        for i in range(0, len(value_list)):
            label = label_list[i].text
            
            result[label] = value_list[i].text

        return result

    def fees(self):
        """
        Returns:
            dict fees
        """
        result = {}
        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=5"
        response = requests.get(url, headers=headers)
        soup = BeautifulSoup(response.text, 'html.parser')
        if soup.find(id='managementFeesDiv') == None:
            return {}
        #label
        label_list =soup.find(id='managementFeesDiv').find_all('td', {"class": "label"})
        #value
        value_list = soup.find(id='managementFeesDiv').find_all('td', {"class": "value number"}) + soup.find(id='managementFeesDiv').find_all('td', {"class": "value number jdpa"})
        for i in range(0, len(value_list)):
            label = label_list[i].text
            result[label] = re.sub('(\\n +)|(\\n)','',value_list[i].text)

        return result


    def searchField(self,pattern = ''):
        """
        Args:
        pattern (str) : text contained in the field

        Returns:
            list of fields possible for the method dataPoint

        Example:
            >>> MS.searchField('fees')
            >>> MS.searchField('return')

        """
        
        regex = re.compile(f'(?i){pattern}')
        filtered_list = list(filter(lambda field : regex.search(field),FIELDS))
        print(f"possible fields for DataPoint can be : {', '.join(filtered_list)} ")
        return filtered_list

    

    def dataPoint(self, field, currency ='EUR' ):
        """
        Args:
        field (str or list) : field to find
        currency (str) : currency in 3 letters

        Returns:
            dict funds infos

        Example:
            >>> MS.dataPoint(['largestSector', 'Name', 'ongoingCharge'])
            >>> MS.dataPoint('SharpeM36')

        """
        return search_funds(self.code, field,self.country,10,currency)
        


    
    

    

    

